from pyproj import CRS

EPSG4326_CRS = CRS.from_epsg(4326)
ARCTIC_POLAR_STEROGRAPHIC_CRS = CRS.from_epsg(3995)
ANTARCTIC_POLAR_STEROGRAPHIC_CRS = CRS.from_epsg(3031)
SOUTH_POLE_CRS = CRS.from_epsg(3031)
EPSILON = 1e-8
